#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os
from typing import Dict, Union, TypeVar, List, Optional, Any, Type
import importlib
import json

from dataclasses_jsonschema import JsonSchemaValidationError, JsonSchemaMixin

from arcor2.object_types_utils import built_in_types_names
from arcor2.data.common import Project, Scene, CurrentAction, Pose, Orientation
from arcor2.data.execution import PackageInfo
from arcor2.data.object_type import ObjectModel, Models, Box, Sphere, Cylinder, Mesh
from arcor2.data.events import CurrentActionEvent, PackageInfoEvent
from arcor2.exceptions import ResourcesException, Arcor2Exception
import arcor2.object_types
from arcor2.object_types import Generic
from arcor2.services import Service, RobotService
from arcor2.action import print_event
from arcor2.rest import convert_keys
from arcor2.helpers import camel_case_to_snake_case, make_pose_abs, make_orientation_abs, print_exception
import arcor2.object_types_utils as otu

from arcor2.parameter_plugins import PARAM_PLUGINS
from arcor2.parameter_plugins.base import TypesDict


# TODO for bound methods - check whether provided action point belongs to the object

class IntResources:

    CUSTOM_OBJECT_TYPES_MODULE = "object_types"
    SERVICES_MODULE = "services"

    def __init__(self, scene: Scene, project: Project, models: Dict[str, Optional[Models]]) -> None:

        self.project = project
        self.scene = scene

        if self.project.scene_id != self.scene.id:
            raise ResourcesException("Project/scene not consistent!")

        self.services: Dict[str, Service] = {}
        self.objects: Dict[str, Generic] = {}

        self.robot_service: Optional[RobotService] = None

        self.type_defs: TypesDict = {}

        for srv in self.scene.services:

            assert srv.type not in self.services, "Duplicate service {}!".format(srv.type)

            module = importlib.import_module(ResourcesBase.SERVICES_MODULE + "." + camel_case_to_snake_case(srv.type))
            cls = getattr(module, srv.type)
            assert issubclass(cls, Service)

            srv_inst = cls(srv.configuration_id)
            self.type_defs[cls.__name__] = cls
            self.services[srv.type] = srv_inst

            if isinstance(srv_inst, RobotService):
                assert self.robot_service is None
                self.robot_service = srv_inst

        built_in = built_in_types_names()

        if self.robot_service:
            self.robot_service.clear_collisions()

        package_info_event = PackageInfoEvent()
        # TODO read package id from package.json
        package_info_event.data = PackageInfo(os.path.basename(os.getcwd()), self.scene, self.project)

        for scene_obj in self.scene.objects:

            if scene_obj.type in built_in:
                module = importlib.import_module(arcor2.object_types.__name__ + "." +
                                                 camel_case_to_snake_case(scene_obj.type))
            else:
                module = importlib.import_module(ResourcesBase.CUSTOM_OBJECT_TYPES_MODULE + "." +
                                                 camel_case_to_snake_case(scene_obj.type))

            cls = getattr(module, scene_obj.type)
            self.type_defs[cls.__name__] = cls

            assert scene_obj.id not in self.objects, "Duplicate object id {}!".format(scene_obj.id)

            if hasattr(cls, "from_services"):

                obj_meta = otu.meta_from_def(cls)

                args: List[Service] = []

                for srv_type in obj_meta.needs_services:
                    args.append(self.services[srv_type])

                # TODO type hints
                inst = cls(*args, scene_obj.id, scene_obj.name, scene_obj.pose, models[scene_obj.type])
            else:
                # TODO type hints
                inst = cls(scene_obj.id, scene_obj.name, scene_obj.pose, models[scene_obj.type])

            self.objects[scene_obj.id] = inst

            if self.robot_service:
                self.robot_service.add_collision(inst)

        if self.robot_service:

            for model in models.values():

                if not model:
                    continue

                if isinstance(model, Box):
                    package_info_event.data.collision_models.boxes.append(model)
                elif isinstance(model, Sphere):
                    package_info_event.data.collision_models.spheres.append(model)
                elif isinstance(model, Cylinder):
                    package_info_event.data.collision_models.cylinders.append(model)
                elif isinstance(model, Mesh):
                    package_info_event.data.collision_models.meshes.append(model)

        print_event(package_info_event)

        self.all_instances: Dict[str, Union[Generic, Service]] = dict(**self.objects, **self.services)

    def make_all_poses_absolute(self) -> None:
        """
        This is only needed when the main script is written fully manually (actions are not defined in AR Editor).
        :return:
        """

        # make all poses absolute
        for aps in self.project.action_points_with_parent:

            assert aps.parent

            obj_inst = self.objects[aps.parent]

            # Action point pose is relative to its parent object pose in scene but is absolute during runtime.
            obj_inst.action_points[aps.id] = aps
            aps.position = make_pose_abs(obj_inst.pose, Pose(aps.position, Orientation())).position
            for ori in aps.orientations:
                ori.orientation = make_orientation_abs(obj_inst.pose.orientation, ori.orientation)

            aps.parent = None

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, traceback):

        if ex_type:  # TODO ignore when script is stopped correctly (e.g. KeyboardInterrupt, ??)
            print_exception(ex_type(ex_value))

        if self.robot_service:
            try:
                for obj in self.objects.values():
                    self.robot_service.remove_collision(obj)
            except Arcor2Exception as e:
                print_exception(e)

        return True

    def print_info(self, action_id: str, args: Dict[str, Any]) -> None:
        """Helper method used to print out info about the action going to be executed."""
        # TODO to be used for parameters that are result of previous action(s)
        # ...there is no need to send parameters that are already in the project
        print_event(CurrentActionEvent(data=CurrentAction(action_id)))

    def parameters(self, action_id: str) -> Dict[str, Any]:

        try:
            act = self.project.action(action_id)
        except Arcor2Exception:
            raise ResourcesException("Action_id {} not found in project {}.".format(action_id, self.project.id))

        ret: Dict[str, Any] = {}

        for param in act.parameters:
            ret[param.id] = PARAM_PLUGINS[param.type].execution_value(self.type_defs, self.scene, self.project,
                                                                      action_id, param.id)

        return ret


T = TypeVar('T', bound=JsonSchemaMixin)


def lower(s: str) -> str:

    return s.lower()


class ResourcesBase(IntResources):

    def read_project_data(self, file_name: str, cls: Type[T]) -> T:

        try:

            with open(os.path.join("data", file_name + ".json")) as scene_file:

                data_dict = json.loads(scene_file.read())
                data_dict = convert_keys(data_dict, camel_case_to_snake_case)

                return cls.from_dict(data_dict)

        except JsonSchemaValidationError as e:
            raise ResourcesException(f"Invalid project/scene: {e}")

    def __init__(self, project_id: str) -> None:

        scene = self.read_project_data(Scene.__name__.lower(), Scene)
        project = self.read_project_data(Project.__name__.lower(), Project)

        if project_id != project.id:
            raise ResourcesException("Resources were generated for different project!")

        models: Dict[str, Optional[Models]] = {}

        for obj in scene.objects:

            if obj.type in models:
                continue

            try:
                models[obj.type] = self.read_project_data(camel_case_to_snake_case(obj.type), ObjectModel).model()
            except IOError:
                models[obj.type] = None

        super(ResourcesBase, self).__init__(scene, project, models)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import json
import logging
import os
import shutil
import tempfile
from typing import Set
from datetime import datetime, timezone

import horast
from apispec import APISpec  # type: ignore
from apispec_webframeworks.flask import FlaskPlugin  # type: ignore
from flask import Flask, send_file, request
from flask_cors import CORS  # type: ignore
from flask_swagger_ui import get_swaggerui_blueprint  # type: ignore

import arcor2
from arcor2 import persistent_storage as ps
from arcor2.data.object_type import ObjectModel
from arcor2.data.execution import PackageMeta
from arcor2.helpers import camel_case_to_snake_case, logger_formatter
from arcor2.object_types_utils import built_in_types_names
from arcor2.source import SourceException
from arcor2.source.logic import program_src  # , get_logic_from_source
from arcor2.source.utils import derived_resources_class, global_action_points_class, global_actions_class

PORT = 5008
SERVICE_NAME = "ARCOR2 Build Service"

logger = logging.getLogger("build")
ch = logging.StreamHandler()
ch.setFormatter(logger_formatter())
logger.setLevel(logging.INFO)
logger.addHandler(ch)

# Create an APISpec
spec = APISpec(
    title=SERVICE_NAME,
    version=arcor2.version(),
    openapi_version="3.0.2",
    plugins=[FlaskPlugin()],
)

app = Flask(__name__)
CORS(app)


def _publish(project_id: str, package_name: str):

    with tempfile.TemporaryDirectory() as tmpdirname:

        try:
            project = ps.get_project(project_id)
            scene = ps.get_scene(project.scene_id)

            project_dir = os.path.join(tmpdirname, "arcor2_project")

            data_path = os.path.join(project_dir, "data")
            ot_path = os.path.join(project_dir, "object_types")
            srv_path = os.path.join(project_dir, "services")

            os.makedirs(data_path)
            os.makedirs(ot_path)
            os.makedirs(srv_path)

            with open(os.path.join(ot_path, "__init__.py"), "w"):
                pass

            with open(os.path.join(srv_path, "__init__.py"), "w"):
                pass

            with open(os.path.join(data_path, "project.json"), "w") as project_file:
                project_file.write(project.to_json())

            with open(os.path.join(data_path, "scene.json"), "w") as scene_file:
                scene_file.write(scene.to_json())

            obj_types_with_models: Set[str] = set()

            for scene_obj in scene.objects:  # TODO handle inheritance

                obj_type = ps.get_object_type(scene_obj.type)

                if obj_type.model and obj_type.id not in obj_types_with_models:
                    obj_types_with_models.add(obj_type.id)

                    model = ps.get_model(obj_type.model.id, obj_type.model.type)
                    obj_model = ObjectModel(obj_type.model.type, **{model.type().value.lower(): model})

                    with open(os.path.join(data_path, camel_case_to_snake_case(obj_type.id) + ".json"), "w")\
                            as model_file:
                        model_file.write(obj_model.to_json())

                with open(os.path.join(ot_path, camel_case_to_snake_case(obj_type.id)) + ".py", "w") as obj_file:
                    obj_file.write(obj_type.source)

            for scene_srv in scene.services:
                srv = ps.get_service_type(scene_srv.type)
                with open(os.path.join(srv_path, camel_case_to_snake_case(srv.id)) + ".py", "w") as srv_file:
                    srv_file.write(srv.source)

        except ps.PersistentStorageException as e:
            logger.exception("Failed to get something from the project service.")
            return str(e), 404

        try:

            with open(os.path.join(project_dir, 'script.py'), "w") as script_file:

                if project.has_logic:
                    script_file.write(program_src(project, scene, built_in_types_names(), True))
                else:
                    try:
                        script = ps.get_project_sources(project.id).script

                        # check if it is a valid Python code
                        try:
                            horast.parse(script)
                        except SyntaxError:
                            logger.exception("Failed to parse code of the uploaded script.")
                            return "Invalid code.", 501

                        script_file.write(script)

                    except ps.PersistentStorageException:

                        logger.info("Script not found on project service, creating one from scratch.")

                        # write script without the main loop
                        script_file.write(program_src(project, scene, built_in_types_names(), False))

            with open(os.path.join(project_dir, 'resources.py'), "w") as res:
                res.write(derived_resources_class(project))

            with open(os.path.join(project_dir, 'actions.py'), "w") as act:
                act.write(global_actions_class(project))

            with open(os.path.join(project_dir, 'action_points.py'), "w") as aps:
                aps.write(global_action_points_class(project))

            with open(os.path.join(project_dir, 'package.json'), "w") as pkg:
                pkg.write(PackageMeta(package_name, datetime.now(tz=timezone.utc)).to_json())

        except SourceException as e:
            logger.exception("Failed to generate script.")
            return str(e), 501

        archive_path = os.path.join(tmpdirname, "arcor2_project")
        shutil.make_archive(archive_path, 'zip',  project_dir)
        return send_file(archive_path + ".zip", as_attachment=True, cache_timeout=0)


@app.route("/project/<string:project_id>/publish", methods=['GET'])
def project_publish(project_id: str):
    """Publish project
            ---
            get:
              description: Get zip file with execution package. To be used by the Execution service.
              parameters:
                - in: path
                  name: project_id
                  schema:
                    type: string
                  required: true
                  description: unique ID
                - in: query
                  name: PackageName
                  schema:
                    type: string
                    default: Package
                  required: false
                  description: Package name
              responses:
                200:
                  description: Ok
                  content:
                    application/zip:
                        schema:
                          type: string
                          format: binary
                          example: The archive of execution package (.zip)
                404:
                    description: Project ID or some of the required items was not found.
                501:
                    description: Project invalid.
            """

    return _publish(project_id, request.args.get('PackageName', default="N/A"))


@app.route("/project/<string:project_id>/script", methods=['PUT'])
def project_script(project_id: str):
    """Project script
            ---
            put:
              description: Add or update project main script (DOES NOT WORK YET).
              parameters:
                - in: path
                  name: project_id
                  schema:
                    type: string
                  required: true
                  description: unique ID
              requestBody:
                  content:
                    text/x-python:
                      schema:
                        type: string
                        format: binary
              responses:
                200:
                  description: Ok
            """
    # TODO use get_logic_from_source
    pass


@app.route("/swagger/api/swagger.json", methods=["GET"])
def get_swagger():
    return json.dumps(spec.to_dict())


with app.test_request_context():
    spec.path(view=project_publish)
    spec.path(view=project_script)


def main():

    parser = argparse.ArgumentParser(description=SERVICE_NAME)
    parser.add_argument('-s', '--swagger', action="store_true", default=False)
    args = parser.parse_args()

    if args.swagger:
        print(spec.to_yaml())
        return

    SWAGGER_URL = "/swagger"

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL,  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
        "./api/swagger.json"
    )

    # Register blueprint at URL
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    app.run(host='0.0.0.0', port=PORT)


if __name__ == '__main__':
    main()



import abc
from typing import Dict, Optional

from arcor2 import DynamicParamDict, CancelDict
from arcor2.data.common import Pose, ActionPoint, ActionMetadata, SceneObject
from arcor2.action import action
from arcor2.data.object_type import Models
from arcor2.docstring import parse_docstring
from arcor2.exceptions import Arcor2Exception


class GenericException(Arcor2Exception):
    pass


class Generic(metaclass=abc.ABCMeta):
    """
    Generic object
    """

    DYNAMIC_PARAMS: DynamicParamDict = {}
    CANCEL_MAPPING: CancelDict = {}

    def __init__(self, obj_id: str, name: str, pose: Pose, collision_model: Optional[Models] = None) -> None:

        self.id = obj_id
        self.name = name
        self.pose = pose
        self.collision_model = collision_model
        self.action_points: Dict[str, ActionPoint] = {}
        self._int_dict: Dict[str, int] = {}

    @classmethod
    def description(cls) -> str:  # TODO mixin with common stuff for objects/services?
        return parse_docstring(cls.__doc__)["short_description"]

    def scene_object(self) -> SceneObject:
        return SceneObject(self.id, self.name, self.__class__.__name__, self.pose)

    def __repr__(self) -> str:
        return str(self.__dict__)

    # TODO allow to store any value?
    @action
    def set_int(self, key: str, value: int) -> None:
        self._int_dict[key] = value

    @action
    def get_int(self, key: str) -> int:
        try:
            return self._int_dict[key]
        except KeyError:
            raise GenericException(f"Unknown key: {key}.")

    set_int.__action__ = ActionMetadata()  # type: ignore
    get_int.__action__ = ActionMetadata()  # type: ignore



import asyncio
from typing import Optional, List, get_type_hints, Set, AsyncIterator, Union

from arcor2.data.common import SceneService, SceneObject, Scene
from arcor2.data import events
from arcor2.data.object_type import Models
from arcor2.exceptions import Arcor2Exception
from arcor2.services import Service, RobotService
from arcor2.object_types import Generic
from arcor2 import aio_persistent_storage as storage, helpers as hlp, object_types_utils as otu

from arcor2.server import globals as glob, notifications as notif, objects_services_actions as osa
from arcor2.server.robot import collision


def instances_names() -> Set[str]:
    return {obj.name for obj in glob.SCENE_OBJECT_INSTANCES.values()}


async def scenes() -> AsyncIterator[Scene]:

    for scene_id in (await storage.get_scenes()).items:
        yield await storage.get_scene(scene_id.id)


async def add_service_to_scene(srv: SceneService, dry_run: bool = False) -> None:

    if srv.type not in glob.SERVICE_TYPES:
        raise Arcor2Exception("Unknown service type.")

    if glob.SERVICE_TYPES[srv.type].disabled:
        raise Arcor2Exception("Service is disabled.")

    if srv.type in glob.SERVICES_INSTANCES:
        raise Arcor2Exception("Service already in scene.")

    if dry_run:
        return None

    srv_type = await storage.get_service_type(srv.type)
    cls_def = hlp.type_def_from_source(srv_type.source, srv_type.id, Service)

    if issubclass(cls_def, RobotService) and osa.find_robot_service():
        raise Arcor2Exception("Scene might contain only one robot service.")

    srv_inst = await hlp.run_in_executor(cls_def, srv.configuration_id)

    glob.SERVICES_INSTANCES[srv.type] = srv_inst

    if isinstance(srv_inst, RobotService):
        await glob.logger.info("RobotService added, adding collision models to all objects.")
        for obj_inst in glob.SCENE_OBJECT_INSTANCES.values():
            await collision(obj_inst, srv_inst, add=True)

    return None


async def add_object_to_scene(obj: SceneObject, add_to_scene=True, srv_obj_ok=False, dry_run: bool = False):
    """

    :param obj:
    :param add_to_scene: Set to false to only create object instance and add its collision model (if any).
    :return:
    """

    assert glob.SCENE

    if obj.type not in glob.OBJECT_TYPES:
        raise Arcor2Exception("Unknown object type.")

    obj_meta = glob.OBJECT_TYPES[obj.type]

    if obj_meta.disabled:
        raise Arcor2Exception("Object type disabled.")

    if srv_obj_ok:  # just for internal usage

        if not obj_meta.needs_services <= osa.valid_service_types().keys():
            raise Arcor2Exception("Some of required services is not available.")

        if not obj_meta.needs_services <= glob.SERVICES_INSTANCES.keys():
            raise Arcor2Exception("Some of required services is not in the scene.")

    elif obj_meta.needs_services:
        raise Arcor2Exception("Service(s)-based object.")

    if obj_meta.abstract:
        raise Arcor2Exception("Cannot instantiate abstract type.")

    if obj.id in glob.SCENE_OBJECT_INSTANCES or obj.id in glob.SERVICES_INSTANCES:
        raise Arcor2Exception("Object/service with that id already exists.")

    if obj.name in instances_names():
        raise Arcor2Exception("Name is already used.")

    if not hlp.is_valid_identifier(obj.name):
        raise Arcor2Exception("Object name invalid (should be snake_case).")

    if dry_run:
        return None

    await glob.logger.debug(f"Creating instance {obj.id} ({obj.type}).")

    if obj.type in otu.built_in_types_names():
        cls = otu.get_built_in_type(obj.type)
    else:
        obj_type = await storage.get_object_type(obj.type)
        cls = hlp.type_def_from_source(obj_type.source, obj_type.id, Generic)

    coll_model: Optional[Models] = None
    if obj_meta.object_model:
        coll_model = obj_meta.object_model.model()

    if not obj_meta.needs_services:
        obj_inst = cls(obj.id, obj.name, obj.pose, coll_model)
    else:

        srv_args: List[Service] = []

        for name, ttype in get_type_hints(cls.__init__).items():

            # service arguments should be listed first
            if not issubclass(ttype, Service):
                break

            try:
                srv_args.append(glob.SERVICES_INSTANCES[ttype.__name__])
            except KeyError:
                return False, f"Object type {obj.type} has invalid typ annotation in the constructor, " \
                              f"service {ttype.__name__} not available."

        try:
            obj_inst = cls(*srv_args, obj.id, obj.name, obj.pose, coll_model)  # type: ignore
        except TypeError as e:
            raise Arcor2Exception("System error.") from e

    glob.SCENE_OBJECT_INSTANCES[obj.id] = obj_inst

    if add_to_scene:
        glob.SCENE.objects.append(obj)

    await collision(obj_inst, add=True)

    return None


async def auto_add_object_to_scene(obj_type_name: str, dry_run: bool = False) -> None:

    assert glob.SCENE

    if obj_type_name not in glob.OBJECT_TYPES:
        raise Arcor2Exception("Unknown object type.")

    if obj_type_name in otu.built_in_types_names():
        raise Arcor2Exception("Does not work for built in types.")

    obj_meta = glob.OBJECT_TYPES[obj_type_name]

    if not obj_meta.needs_services:
        raise Arcor2Exception("Ordinary object.")

    if obj_meta.abstract:
        raise Arcor2Exception("Cannot instantiate abstract type.")

    if not obj_meta.needs_services <= osa.valid_service_types().keys():
        raise Arcor2Exception("Some of required services is not available.")

    if not obj_meta.needs_services <= glob.SERVICES_INSTANCES.keys():
        raise Arcor2Exception("Some of required services is not in the scene.")

    if dry_run:
        return None

    obj_type = await storage.get_object_type(obj_type_name)
    cls = hlp.type_def_from_source(obj_type.source, obj_type.id, Generic)

    args: List[Service] = [glob.SERVICES_INSTANCES[srv_name] for srv_name in obj_meta.needs_services]

    assert hasattr(cls, otu.SERVICES_METHOD_NAME)
    for obj_inst in cls.from_services(*args):  # type: ignore

        assert isinstance(obj_inst, Generic)

        if not hlp.is_valid_identifier(obj_inst.name):
            # TODO add message to response
            await glob.logger.warning(f"Object id {obj_inst.id} invalid.")
            continue

        if obj_inst.id in glob.SCENE_OBJECT_INSTANCES:
            await glob.logger.warning(f"Object id {obj_inst.id} already in scene.")
            continue

        if obj_inst.name in instances_names():
            await glob.logger.warning(f"Duplicate object name {obj_inst.name}.")
            continue

        glob.SCENE_OBJECT_INSTANCES[obj_inst.id] = obj_inst

        obj = obj_inst.scene_object()
        glob.SCENE.objects.append(obj)

        asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.ADD, data=obj)))

        if obj_meta.object_model:
            obj_inst.collision_model = obj_meta.object_model.model()
            await collision(obj_inst, add=True)

    return None


async def clear_scene() -> None:

    await glob.logger.info("Clearing the scene.")
    rs = osa.find_robot_service()
    if rs:
        for obj_inst in glob.SCENE_OBJECT_INSTANCES.values():
            await collision(obj_inst, rs, remove=True)
    glob.SCENE_OBJECT_INSTANCES.clear()
    glob.SERVICES_INSTANCES.clear()  # TODO call destroy

    glob.SCENE = None


async def open_scene(scene_id: str) -> None:

    glob.SCENE = await storage.get_scene(scene_id)

    try:

        for srv in glob.SCENE.services:
            await add_service_to_scene(srv)

        for obj in glob.SCENE.objects:
            await add_object_to_scene(obj, add_to_scene=False, srv_obj_ok=True)

    except Arcor2Exception as e:
        await clear_scene()
        raise Arcor2Exception(f"Failed to open scene. {e.message}") from e

    assert {srv.type for srv in glob.SCENE.services} == glob.SERVICES_INSTANCES.keys()
    assert {obj.id for obj in glob.SCENE.objects} == glob.SCENE_OBJECT_INSTANCES.keys()


def get_instance(obj_id: str) -> Union[Generic, Service]:

    if obj_id in glob.SCENE_OBJECT_INSTANCES:
        return glob.SCENE_OBJECT_INSTANCES[obj_id]
    elif obj_id in glob.SERVICES_INSTANCES:
        return glob.SERVICES_INSTANCES[obj_id]
    else:
        raise Arcor2Exception("Unknown object/service ID.")



import abc
from typing import FrozenSet
from arcor2 import DynamicParamDict, CancelDict
from arcor2.docstring import parse_docstring


class Service(metaclass=abc.ABCMeta):
    """
    A generic service.
    """

    DYNAMIC_PARAMS: DynamicParamDict = {}
    CANCEL_MAPPING: CancelDict = {}

    def __init__(self, configuration_id: str) -> None:
        self.configuration_id = configuration_id

    @property
    def id(self) -> str:
        return self.__class__.__name__

    @classmethod
    def description(cls) -> str:
        return parse_docstring(cls.__doc__)["short_description"]

    @staticmethod
    def get_configuration_ids() -> FrozenSet[str]:
        return frozenset()



from setuptools import setup, find_packages  # type: ignore
import arcor2

setup(
    name='arcor2',
    version=arcor2.version(),
    include_package_data=True,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_data={"arcor2": ["py.typed"]},
    entry_points={
                  'console_scripts': [
                      'arcor2_server = arcor2.nodes.server:main',
                      'arcor2_execution = arcor2.nodes.execution:main',
                      'arcor2_build = arcor2.nodes.build:main',
                      'arcor2_upload_ot = arcor2.scripts.upload_object_type:main',
                      'arcor2_swagger_models = arcor2.scripts.swagger_models:main',
                      'arcor2_upload_services = arcor2.scripts.upload_services:main',
                      'arcor2_execution_proxy = arcor2.nodes.execution_proxy:main',
                      'arcor2_broadcaster = arcor2.nodes.broadcaster:main'
                  ],
              },
    url='https://github.com/robofit/arcor2',
    download_url=f'https://github.com/robofit/arcor2/archive/{arcor2.version()}.tar.gz',
    license='LGPL',
    author='Robo@FIT',
    author_email='imaterna@fit.vutbr.cz',
    description='',
    install_requires=[
        'dataclasses',
        'horast',
        'typed_ast',
        'autopep8',
        'typed_astunparse',
        'static_typing',
        'websockets',
        'aiologger',
        'aiofiles',
        'dataclasses-jsonschema[fast-validation]',
        'apispec',
        'apispec_webframeworks',
        'flask',
        'requests',
        'cython',  # dependency of numpy, for some reason not installed automatically...
        'numpy-quaternion',
        'fastcache',
        'flask_swagger_ui',
        'websocket-client',
        'pyyaml',  # dependency of apispec, for some reason not installed automatically...
        'Pillow',
        'aiorun',
        'flask-cors'
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-docker-compose',
            'openapi-spec-validator'
            ],
        'docs': ['sphinx']
    },
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.8',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering'
    ]
)



import abc
from typing import List, FrozenSet

from arcor2.services import Service
from arcor2.data.common import Pose, Joint
from arcor2.data.object_type import MeshFocusAction
from arcor2.object_types import Generic


class RobotService(Service, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def add_collision(self, obj: Generic) -> None:
        """
        Adds collision model for object instance.
        :param obj:
        :return:
        """
        pass

    @abc.abstractmethod
    def clear_collisions(self):
        pass

    @abc.abstractmethod
    def remove_collision(self, obj: Generic) -> None:
        """
        Removes collision model for object instance.
        :param obj:
        :return:
        """
        pass

    @abc.abstractmethod
    def get_robot_ids(self) -> FrozenSet[str]:
        pass

    @abc.abstractmethod
    def get_robot_pose(self, robot_id: str) -> Pose:
        pass

    def stop(self, robot_id: str) -> None:
        pass

    @abc.abstractmethod
    def get_end_effectors_ids(self, robot_id: str) -> FrozenSet[str]:
        pass

    @abc.abstractmethod
    def get_end_effector_pose(self, robot_id: str, end_effector_id: str) -> Pose:
        pass

    @abc.abstractmethod
    def move(self, robot_id: str, end_effector_id: str, pose: Pose):
        pass

    @abc.abstractmethod
    def focus(self, mfa: MeshFocusAction) -> Pose:
        pass

    def inputs(self, robot_id: str) -> FrozenSet[str]:
        return frozenset()

    def outputs(self, robot_id: str) -> FrozenSet[str]:
        return frozenset()

    def get_input(self, robot_id: str, input_id: str) -> float:
        if input_id not in self.inputs(robot_id):
            raise ValueError("Invalid input_id.")

        return 0

    def set_output(self, robot_id: str, output_id: str, value: float) -> None:
        if output_id not in self.outputs(robot_id):
            raise ValueError("Invalid output_id.")
        return None

    @abc.abstractmethod
    def grippers(self, robot_id: str) -> FrozenSet[str]:
        return frozenset()

    @abc.abstractmethod
    def suctions(self, robot_id: str) -> FrozenSet[str]:
        return frozenset()

    @abc.abstractmethod
    def robot_joints(self, robot_id: str) -> List[Joint]:
        pass



import asyncio
import time
from typing import Dict, Callable, Awaitable

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.data import rpc, events
from arcor2.exceptions import Arcor2Exception
from arcor2 import helpers as hlp

from arcor2.server import globals as glob, objects_services_actions as osa, robot
from arcor2.server.decorators import scene_needed

TaskDict = Dict[str, asyncio.Task]

ROBOT_JOINTS_TASKS: TaskDict = {}
EEF_POSE_TASKS: TaskDict = {}

EVENT_PERIOD = 0.1


async def robot_joints_event(robot_id: str) -> None:

    global ROBOT_JOINTS_TASKS

    await glob.logger.info(f"Sending '{events.RobotJointsEvent.__name__}' for robot '{robot_id}' started.")
    while glob.SCENE and glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]:

        start = time.monotonic()

        evt = events.RobotJointsEvent()

        try:
            evt.data = events.RobotJointsData(robot_id, (await robot.get_robot_joints(robot_id)))
        except Arcor2Exception as e:
            glob.logger.error(f"Failed to get joints for {robot_id}. {e.message}")
            break

        evt_json = evt.to_json()
        await asyncio.gather(*[hlp.send_json_to_client(ui, evt_json)
                               for ui in glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]])

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD-(end-start))

    del ROBOT_JOINTS_TASKS[robot_id]

    # TODO notify UIs that registration was cancelled
    del glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]

    await glob.logger.info(f"Sending '{events.RobotJointsEvent.__name__}' for robot '{robot_id}' stopped.")


async def eef_pose(robot_id: str, eef_id: str) -> events.EefPose:

    return events.EefPose(eef_id, (await robot.get_end_effector_pose(robot_id, eef_id)))


async def robot_eef_pose_event(robot_id: str) -> None:

    global EEF_POSE_TASKS

    await glob.logger.info(f"Sending '{events.RobotEefEvent.__name__}' for robot '{robot_id}' started.")

    while glob.SCENE and glob.ROBOT_EEF_REGISTERED_UIS[robot_id]:

        start = time.monotonic()

        evt = events.RobotEefEvent()
        evt.data = events.RobotEefData(robot_id)

        try:
            evt.data.end_effectors = await asyncio.gather(
                *[eef_pose(robot_id, eef_id) for eef_id in (await robot.get_end_effectors(robot_id))])
        except Arcor2Exception as e:
            glob.logger.error(f"Failed to get eef pose for {robot_id}. {e.message}")
            break

        evt_json = evt.to_json()
        await asyncio.gather(*[hlp.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_EEF_REGISTERED_UIS[robot_id]])

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD-(end-start))

    del EEF_POSE_TASKS[robot_id]

    # TODO notify UIs that registration was cancelled
    del glob.ROBOT_EEF_REGISTERED_UIS[robot_id]

    await glob.logger.info(f"Sending '{events.RobotEefEvent.__name__}' for robot '{robot_id}' stopped.")


async def get_robot_meta_cb(req: rpc.robot.GetRobotMetaRequest, ui: WsClient) ->\
        rpc.robot.GetRobotMetaResponse:

    return rpc.robot.GetRobotMetaResponse(data=list(glob.ROBOT_META.values()))


@scene_needed
async def get_robot_joints_cb(req: rpc.robot.GetRobotJointsRequest, ui: WsClient) -> \
        rpc.robot.GetRobotJointsResponse:

    return rpc.robot.GetRobotJointsResponse(data=await robot.get_robot_joints(req.args.robot_id))


@scene_needed
async def get_end_effector_pose_cb(req: rpc.robot.GetEndEffectorPoseRequest, ui: WsClient) -> \
        rpc.robot.GetEndEffectorPoseResponse:

    return rpc.robot.GetEndEffectorPoseResponse(
        data=await robot.get_end_effector_pose(req.args.robot_id, req.args.end_effector_id))


@scene_needed
async def get_end_effectors_cb(req: rpc.robot.GetEndEffectorsRequest, ui: WsClient) -> \
        rpc.robot.GetEndEffectorsResponse:

    return rpc.robot.GetEndEffectorsResponse(data=await robot.get_end_effectors(req.args.robot_id))


@scene_needed
async def get_grippers_cb(req: rpc.robot.GetGrippersRequest, ui: WsClient) -> \
        rpc.robot.GetGrippersResponse:

    return rpc.robot.GetGrippersResponse(data=await robot.get_grippers(req.args.robot_id))


@scene_needed
async def get_suctions_cb(req: rpc.robot.GetSuctionsRequest, ui: WsClient) -> \
        rpc.robot.GetSuctionsResponse:

    return rpc.robot.GetSuctionsResponse(data=await robot.get_suctions(req.args.robot_id))


async def register(req: rpc.robot.RegisterForRobotEventRequest, ui: WsClient, tasks: TaskDict,
                   reg_uis: glob.RegisteredUiDict, coro: Callable[[str], Awaitable[None]]) -> None:

    if req.args.send:

        reg_uis[req.args.robot_id].add(ui)

        if req.args.robot_id not in tasks:
            # start task
            tasks[req.args.robot_id] = asyncio.create_task(coro(req.args.robot_id))

    else:
        try:
            reg_uis[req.args.robot_id].remove(ui)
        except KeyError as e:
            raise Arcor2Exception("Failed to unregister.") from e

        # cancel task if not needed anymore
        if not reg_uis[req.args.robot_id]:
            task = tasks[req.args.robot_id]

            if not task.cancelled():
                task.cancel()

            del tasks[req.args.robot_id]


@scene_needed
async def register_for_robot_event_cb(req: rpc.robot.RegisterForRobotEventRequest, ui: WsClient) -> None:

    # check if robot exists
    await osa.get_robot_instance(req.args.robot_id)

    if req.args.what == rpc.robot.RegisterEnum.JOINTS:
        await register(req, ui, ROBOT_JOINTS_TASKS, glob.ROBOT_JOINTS_REGISTERED_UIS,
                       robot_joints_event)
    elif req.args.what == rpc.robot.RegisterEnum.EEF_POSE:

        if not (await robot.get_end_effectors(req.args.robot_id)):
            raise Arcor2Exception("Robot does not have any end effector.")

        await register(req, ui, EEF_POSE_TASKS, glob.ROBOT_EEF_REGISTERED_UIS,
                       robot_eef_pose_event)
    else:
        raise Arcor2Exception(f"Option '{req.args.what.value}' not implemented.")

    return None



from typing import Set, AsyncIterator, List, Dict

from arcor2.data.common import ActionIOEnum, Project, Scene
import arcor2.aio_persistent_storage as storage
from arcor2.parameter_plugins import PARAM_PLUGINS
from arcor2.parameter_plugins.base import ParameterPluginException
from arcor2.exceptions import Arcor2Exception

from arcor2.server import globals as glob
from arcor2.server.scene import open_scene, clear_scene


async def project_names() -> Set[str]:
    return {proj.name for proj in (await storage.get_projects()).items}


async def associated_projects(scene_id: str) -> Set[str]:
    return {project.id async for project in projects(scene_id)}


async def scene_object_pose_updated(scene_id: str, obj_id: str) -> None:
    """
    Invalidates robot joints if action point's parent has changed its pose.
    :param scene_id:
    :param obj_id:
    :return:
    """

    async for project in projects_using_object_as_parent(scene_id, obj_id):

        for ap in project.action_points:

            if ap.parent != obj_id:
                continue

            await glob.logger.debug(f"Invalidating joints for {project.name}/{ap.name}.")
            ap.invalidate_joints()

        await storage.update_project(project)


async def remove_object_references_from_projects(obj_id: str) -> None:

    assert glob.SCENE

    updated_project_ids: Set[str] = set()

    async for project in projects_using_object_as_parent(glob.SCENE.id, obj_id):

        action_ids: Set[str] = set()

        for ap in project.action_points:

            # delete actions using the object
            ap.actions = [act for act in ap.actions if act.parse_type()[0] != obj_id]

            # delete actions using obj's action points as parameters
            # TODO fix this!
            """
            actions_using_invalid_param: Set[str] = \
                {act.id for act in ap.actions for param in act.parameters
                 if param.type in (ActionParameterTypeEnum.JOINTS, ActionParameterTypeEnum.POSE) and
                 param.value.startswith(obj_id)}

            ap.actions = [act for act in ap.actions if act.id not in actions_using_invalid_param]

            # get IDs of remaining actions
            action_ids.update({act.id for act in ap.actions})
            """

        valid_ids: Set[str] = action_ids | ActionIOEnum.set()

        # remove invalid inputs/outputs
        for ap in project.action_points:
            for act in ap.actions:
                act.inputs = [input for input in act.inputs if input.default in valid_ids]
                act.outputs = [output for output in act.outputs if output.default in valid_ids]

        await storage.update_project(project)
        updated_project_ids.add(project.id)

    await glob.logger.info("Updated projects: {}".format(updated_project_ids))


async def projects(scene_id: str) -> AsyncIterator[Project]:

    id_list = await storage.get_projects()

    for project_meta in id_list.items:

        project = await storage.get_project(project_meta.id)

        if project.scene_id != scene_id:
            continue

        yield project


def _project_using_object_as_parent(project: Project, obj_id: str) -> bool:

    for ap in project.action_points_with_parent:

        assert ap.parent

        if ap.parent == obj_id:
            return True

    return False


def _project_referencing_object(project: Project, obj_id: str) -> bool:

    for ap in project.action_points:
        for action in ap.actions:
            action_obj_id, _ = action.parse_type()

            if action_obj_id == obj_id:
                return True

    return False


async def projects_using_object(scene_id: str, obj_id: str) -> AsyncIterator[Project]:
    """
    Combines functionality of projects_using_object_as_parent and projects_referencing_object.
    :param scene_id:
    :param obj_id:
    :return:
    """

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id) or _project_referencing_object(project, obj_id):
            yield project


async def projects_using_object_as_parent(scene_id: str, obj_id: str) -> AsyncIterator[Project]:

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id):
            yield project


async def projects_referencing_object(scene_id: str, obj_id: str) -> AsyncIterator[Project]:

    async for project in projects(scene_id):
        if _project_referencing_object(project, obj_id):
            yield project


def project_problems(scene: Scene, project: Project) -> List[str]:

    scene_objects: Dict[str, str] = {obj.id: obj.type for obj in scene.objects}
    scene_services: Set[str] = {srv.type for srv in scene.services}

    action_ids: Set[str] = set()
    problems: List[str] = []

    unknown_types = ({obj.type for obj in scene.objects} | scene_services) - glob.ACTIONS.keys()

    if unknown_types:
        return [f"Scene invalid, contains unknown types: {unknown_types}."]

    for ap in project.action_points:

        # test if all objects exists in scene
        if ap.parent and ap.parent not in scene_objects:
            problems.append(f"Action point '{ap.name}' has parent '{ap.parent}' that does not exist in the scene.")
            continue

        for joints in ap.robot_joints:
            if not joints.is_valid:
                problems.append(f"Action point {ap.name} has invalid joints: {joints.name} "
                                f"(robot {joints.robot_id}).")

        for action in ap.actions:

            if action.id in action_ids:
                problems.append(f"Action ID {action.name} of the {ap.name} is not unique.")

            # check if objects have used actions
            obj_id, action_type = action.parse_type()

            if obj_id not in scene_objects.keys() | scene_services:
                problems.append(f"Object ID {obj_id} which action is used in {action.name} does not exist in scene.")
                continue

            try:
                os_type = scene_objects[obj_id]  # object type
            except KeyError:
                os_type = obj_id  # service

            for act in glob.ACTIONS[os_type]:
                if action_type == act.name:
                    break
            else:
                problems.append(f"Object type {scene_objects[obj_id]} does not have action {action_type} "
                                f"used in {action.id}.")

            # check object's actions parameters
            action_params: Dict[str, str] = \
                {param.id: param.type for param in action.parameters}
            ot_params: Dict[str, str] = {param.name: param.type for param in act.parameters
                                         for act in glob.ACTIONS[os_type]}

            if action_params != ot_params:
                problems.append(f"Action ID {action.id} of type {action.type} has invalid parameters.")

            # TODO validate parameter values / instances (for value) are not available here / how to solve it?
            for param in action.parameters:
                try:
                    PARAM_PLUGINS[param.type].value(glob.TYPE_DEF_DICT, scene, project, action.id, param.id)
                except ParameterPluginException:
                    problems.append(f"Parameter {param.id} of action {act.name} "
                                    f"has invalid value: '{param.value}'.")

    return problems


async def open_project(project_id: str) -> None:

    project = await storage.get_project(project_id)

    if glob.SCENE:
        if glob.SCENE.id != project.scene_id:
            raise Arcor2Exception("Required project is associated to another scene.")
    else:
        await open_scene(project.scene_id)

    assert glob.SCENE
    for ap in project.action_points_with_parent:

        assert ap.parent
        try:
            glob.SCENE.object(ap.parent)
        except Arcor2Exception:
            await clear_scene()
            raise Arcor2Exception(f"Action point's {ap.name} parent not available in the scene.")

    glob.PROJECT = project



from typing import FrozenSet

from arcor2.services import Service
from arcor2.data.common import ActionMetadata
from arcor2.action import action


class LogicService(Service):
    """
    Logic-related actions.
    """

    def __init__(self, configuration_id: str):  # TODO avoid need for configuration_id?
        super(LogicService, self).__init__(configuration_id)

    @staticmethod
    def get_configuration_ids() -> FrozenSet[str]:
        return frozenset({"default"})

    @action
    def equals(self, val1: int, val2: int) -> bool:
        return val1 == val2

    @action
    def less_than(self, val1: int, val2: int) -> bool:
        return val1 < val2

    @action
    def greater_than(self, val1: int, val2: int) -> bool:
        return val1 > val2

    equals.__action__ = ActionMetadata(blocking=True)  # type: ignore
    less_than.__action__ = ActionMetadata(blocking=True)  # type: ignore
    greater_than.__action__ = ActionMetadata(blocking=True)  # type: ignore



from dataclasses import dataclass, field
from datetime import datetime

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data import common, object_type


@dataclass
class PackageInfo(JsonSchemaMixin):

    package_id: str
    scene: common.Scene
    project: common.Project
    collision_models: object_type.CollisionModels = field(default_factory=object_type.CollisionModels)


@dataclass
class PackageMeta(JsonSchemaMixin):

    name: str
    built: datetime



import functools

from arcor2.server import globals as glob


def no_scene(coro):
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs):

        if glob.SCENE:
            return False, "Scene has to be closed first."
        return await coro(*args, **kwargs)

    return async_wrapper


def scene_needed(coro):
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs):

        if glob.SCENE is None or not glob.SCENE.id:
            return False, "Scene not opened or has invalid id."
        return await coro(*args, **kwargs)

    return async_wrapper


def no_project(coro):
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs):
        if glob.PROJECT:
            return False, "Not available during project editing."
        return await coro(*args, **kwargs)

    return async_wrapper


def project_needed(coro):
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs):

        if glob.PROJECT is None or not glob.PROJECT.id:
            return False, "Project not opened or has invalid id."
        return await coro(*args, **kwargs)

    return async_wrapper



import tempfile
import os
import base64
import uuid


from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.data import rpc
import arcor2.helpers as hlp
from arcor2 import rest
from arcor2.data import common

from arcor2.server.execution import manager_request
import arcor2.server.globals as glob


async def build_project_cb(req: rpc.execution.BuildProjectRequest, ui: WsClient) -> rpc.execution.BuildProjectResponse:
    """
    Builds project and uploads resulting package to the execution unit.
    :param req:
    :param ui:
    :return:
    """

    package_id = common.uid()

    # call build service
    # TODO store data in memory
    with tempfile.TemporaryDirectory() as tmpdirname:

        path = os.path.join(tmpdirname, "publish.zip")

        await hlp.run_in_executor(rest.download, f"{glob.BUILDER_URL}/project/{req.args.project_id}/publish", path,
                                  None, {"package_name": req.args.package_name})

        with open(path, "rb") as zip_file:
            b64_bytes = base64.b64encode(zip_file.read())
            b64_str = b64_bytes.decode()

    # send data to execution service
    exe_req = rpc.execution.UploadPackageRequest(uuid.uuid4().int,
                                                 args=rpc.execution.UploadPackageArgs(package_id, b64_str))
    exe_resp = await manager_request(exe_req)

    resp = rpc.execution.BuildProjectResponse()
    resp.data = rpc.execution.BuildProjectData(package_id)
    resp.result = exe_resp.result
    resp.messages = exe_resp.messages

    return resp



import json

from arcor2.data.common import Project, ProjectRobotJoints, Scene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict, ParameterPluginException
from arcor2.services import RobotService


class JointsPlugin(ParameterPlugin):

    @classmethod
    def type(cls):
        return ProjectRobotJoints

    @classmethod
    def type_name(cls) -> str:
        return "joints"

    @classmethod
    def value(cls, type_defs: TypesDict, scene: Scene, project: Project, action_id: str, parameter_id: str) -> \
            ProjectRobotJoints:

        ap, action = project.action_point_and_action(action_id)
        param = action.parameter(parameter_id)
        joints_id = cls.param_value(param)

        robot_id, action_method_name = action.parse_type()
        robot_type = scene.object_or_service(robot_id)

        if issubclass(type_defs[robot_type.type], RobotService):

            for param in action.parameters:
                if param.id == "robot_id":
                    robot_id = json.loads(param.value)
                    break
            else:
                raise ParameterPluginException(f"Parameter {param.id} of action {action.id} depends on"
                                               f" 'robot_id' parameter, which could not be found.")

        return ap.joints_for_robot(robot_id, joints_id)

    @classmethod
    def value_to_json(cls, value: ProjectRobotJoints) -> str:
        return value.to_json()

    @classmethod
    def uses_robot_joints(cls, project: Project, action_id: str, parameter_id: str, robot_joints_id: str) -> bool:

        param = project.action(action_id).parameter(parameter_id)
        value_id = cls.param_value(param)

        return value_id == robot_joints_id
